//----------------------------------------------------------
           
// common
#include <iostream> 
#include <conio>
#include <exception>

// include the tree root
#include "demo_tree_list.h"
#include "demo_tree_integer.h"

// define explicitly the namespace
namespace demo
{

// direct printing of a zero terminated UTF32 character erray
static void sprintf_utf32(const _uchar *text)
{
  while(*text) std::printf("%c", *text++);       
}            

// direct printing of a UTF32 character erray
static void sprintf_utf32(const _uchar *text, _uint length)
{
  for(_uint i=0; i<length; i++)
   std::printf("%c", text[i]);       
}            

//----------------------------------------------------------
//      CustomEvents extends events
//
// this class receives events during
// the parsing machine runtime
//
//----------------------------------------------------------
class CustomEvents : public events
{
  public:       
    virtual     void    OnSyntaxError(const syntax_error &);
    virtual     void    OnExplored(void);
    virtual     void    OnSuccess(void);
};


// event function called for each syntax error found
void CustomEvents::OnSyntaxError(const syntax_error &e)
{
  // error header
  std::printf("ERROR: ");

  // conert the syntax error do a character array
  _uchar *text = e.ToArray(NULL);

  // print
  sprintf_utf32(text);

  // free the memory
  delete[] text;

  // line termination
  std::printf("\n");  
}

// event function called when exploration has completed
// and there is no syntax tree constructed
void CustomEvents::OnExplored(void)
{
  std::printf("INFO: Explored.\n");
}

// event function called a syntax tree is constructed 
void CustomEvents::OnSuccess(void)
{
  std::printf("INFO: Success.\n");
}

// prepare, run and check the parsing machine completion
// status; returns 'true' on success and 'false' of an
// error occured in which case an error message is printed
static bool Execute(machine &m)
{
  // prepare the machine for running                        
  {            
    _uint status = m.Prepare();
    switch(status)                                           
    {
      case E_OK: break; // the preparation was successfull

      default:
      {
        // failed to prepare the machine, print the error code
        std::printf("ERROR: Preparation error 0x%08X.\n", status);
        return false;
      }
    }
  }

  // run the machine till completion (success or an error)
  {
    _uint status = m.RunST(-1);
    switch(status)
    {                           
      case E_DONE: break; // the running was successfully

      default:
      {
        // there is a runtime error
        std::printf("ERROR: Runtime error 0x%08X.\n", status);
        return false;
      }
    }
  }

  // check the status of the correct machine run
  {
    state_vector sv;
    MACHINE_STATUS status = m.GetState(sv);
    switch(status)
    {
      // a valid syntax tree was constructed
      case MS_COMPLETE:
        break;

      // different no full syntax tree constructed cases
      case MS_WORKING :
        std::printf("ERROR: Invalid working state.\n");
        return false;
      case MS_NEED_INPUT:
        std::printf("ERROR: Completed with a request for more input.\n");
        return false;
      case MS_ERROR:
        std::printf("ERROR: Completed with a syntax error.\n");
        return false;
      case MS_FATAL:
        std::printf("ERROR: Completed with a fatal error.\n");
        return false;
      case MS_INCONSISTENCY:
        std::printf("ERROR: Completed with an inconsistency state.\n");
        return false;
      default:
        std::printf("ERROR: Completed with an unknown state 0x%08X.\n", status);
        return false;
    }
  }

  // having a syntax tree
  return true;
}

static void InvalidTree(void)
{
  throw std::runtime_error("Bad formed syntax tree");
}

static void PrintTree_Integer(demo_integer *i, _uint depth)
{
  // ident based on the depth
  for(_uint k=0; k<depth; k++)
    std::printf(" ");

  // check the integer rule concatenation index
  switch(i->Index())
  {
    case 0:
    {
      // iterate the digits found and print them
      for(demo::list<_uint>::Item *e = i->alt.c0->h0.First;
          e; e = e->Next)
        std::printf("%c", e->Data);
    } break;

    // the integer rule have only one concatenation
    default: InvalidTree();
  }

  // end
  std::printf("\n");
}

static void PrintTree_String(demo::string &s, _uint depth)
{
  // ident based on the depth
  for(_uint k=0; k<depth; k++)
    std::printf(" ");

  // print all chars in the string
  sprintf_utf32(s.Text(), s.Length());

  // end
  std::printf("\n");
}

static void PrintTree(demo_list *list)
{
  // header
  std::printf("TREE:\n");

  // check the concatenation index of the 'list' rule
  switch(list->Index())
  {
    case 0:
    {
      // have the type of the list alternative/concatenation0
      typedef demo_list::ALT::C0 list_C0_t;

      // print the first rule integer 'r0integer'
      // that is the start of the 'list' rule
      PrintTree_Integer(list->alt.c0->r0integer, 0);

      // have the group item type
      typedef demo::list<list_C0_t::A1_GROUP *>::Item group_t;
      
      // iterate all second element that is a list of groups
      for(group_t *e = list->alt.c0->g1.First; e; e = e->Next)
      {
        // entering in depth
        std::printf("(\n");

        // check the concatenation index
        switch(e->Data->Index)
        {
          case 0: // first concatenation 
          {
            // get a pointer to the group
            list_C0_t::A1::C0 *groupC0 = e->Data->Value.c0;

            // print the first item string 's0'
            PrintTree_String(groupC0->s0, 1);

            // print the second item rule integer 'r1integer'
            PrintTree_Integer(groupC0->r1integer, 1);
          } break;

          // the first group does have one concatenation only
          default:
            InvalidTree();
        }                        

        // leaving the depth
        std::printf(")\n");
      }
    } break;

    // the grammar have only a single concatenation inside
    // any other valid is invalid
    default: InvalidTree();
  }
}

// parsing of a byte array  
static void ParseBytes(const _uint8 *input, _uint length)
{
  // create a memory reader stream object
  stream_reader_memory Reader(input, length);

  // create a read buffer object for this stream
  source_buffer Buffer(&Reader, 4096);

  // use the buffer as ASCII
  source_ascii Decoder(&Buffer);

  // create the class that will receive error/success
  // message when the parsing machine runs
  CustomEvents Events;

  // create a memory pool object used by the parsing machine
  memory_pool Pool;

  // create the parsing machine using the memory pool,
  // the stream decoder, the events object and the name
  // of the parsing grammar rule
  machine Machine(&Pool, &Decoder, &Events, "list");

  // execute the machine till an error ocures or a
  // valid syntax tree is constructed
  if (!Execute(Machine))
    return; // no syntax tree was constructed

  // get the result syntax tree
  demo_list *result;
  if (!Machine.GetResult(&result))
  {
    std::printf("ERROR: Unable to get the syntax tree.\n");
    return;
  }

  // use the tree here, currently just print
  // a message for the success
  std::printf("SUCCESS: The syntax tree is available.\n");

  // print the tree
  PrintTree(result);
}

// parse a zero terminated character array
static void Parse(const char *input)
{
  // print a header for the current parsing operation
  std::printf("================================");
  std::printf("\nInput:\n%s\n", input);
  std::printf("--------------------------------");
  std::printf("\nParse Log:\n");

  // do the parsing as raw bytes, using the
  // zero terminated character array
  try
  {
    ParseBytes(reinterpret_cast<const _uint8 *>(input), strlen(input));
  }     
  catch(const std::exception &e)
  {
    // in case of an exception, print it to the console
    std::printf("********************************");
    std::printf("Exception: %s\n", e.what());
  }

  // operation has completed
  std::printf("\n");
}

} // namespace demo                                  
                                              
static void PressAnyKey(void)
{
  // wait key message
  std::printf("\nPress any key to continue...");

  // wait for a single button to be pressed
  _getch();
}

// program entry point
int main(int, char *[])
{
  // check was the parsing machine code compiled
  // with the expected integer type sizes
  try { demo::check_compiled_options(); }
  catch(const std::exception &e)
  {
    std::printf("Exception: %s\n", e.what());
    return -1;
  }

  // do parsing of a correct input
  demo::Parse("54,12,74561");

  // do parsing of a wrong input
  // (because char 'x' is not a number)
  demo::Parse("1,9,x,4");

  // wait for a key
  PressAnyKey();
  return 0;
}

 
